# pylint: disable=E1101
from enum import Enum as EnumType
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union

from fideslang.default_taxonomy import DEFAULT_TAXONOMY
from fideslang.models import DataCategory as FideslangDataCategory
from fideslang.validation import FidesKey
from sqlalchemy import Column
from sqlalchemy import Enum as EnumColumn
from sqlalchemy import ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Session, backref, declared_attr, relationship  # type: ignore
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api import common_exceptions
from fides.api.common_exceptions import (
    StorageConfigNotFoundException,
    WebhookOrderException,
)
from fides.api.db.base_class import Base, FidesBase, JSONTypeOverride
from fides.api.models.client import ClientDetail
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.sql_models import DataCategory  # type: ignore
from fides.api.models.storage import StorageConfig, get_active_default_storage_config
from fides.api.schemas.policy import ActionType, DrpAction
from fides.api.util.data_category import _validate_data_category
from fides.config import CONFIG

if TYPE_CHECKING:
    from fides.api.graph.traversal import TraversalNode


def _validate_drp_action(drp_action: Optional[str]) -> None:
    """Check that DRP action is supported"""
    if not drp_action:
        return
    if drp_action in [
        DrpAction.sale_opt_in.value,
        DrpAction.sale_opt_out.value,
        DrpAction.access_categories.value,
        DrpAction.access_specific.value,
    ]:
        raise common_exceptions.DrpActionValidationError(
            f"{drp_action} action is not supported at this time."
        )


def _validate_rule(
    action_type: Optional[str],
    storage_destination_id: Optional[str],
    masking_strategy: Optional[Dict[str, Union[str, Dict[str, str]]]],
) -> None:
    """Check that the rule's action_type and storage_destination are valid."""
    if not action_type:
        raise common_exceptions.RuleValidationError("action_type is required.")

    if action_type == ActionType.erasure.value:
        if storage_destination_id is not None:
            raise common_exceptions.RuleValidationError(
                "Erasure Rules cannot have storage destinations."
            )
        if masking_strategy is None:
            raise common_exceptions.RuleValidationError(
                "Erasure Rules must have masking strategies."
            )
    if action_type in [ActionType.update.value]:
        raise common_exceptions.RuleValidationError(
            f"{action_type} Rules are not supported at this time."
        )


class Policy(Base):
    """A set of constraints to apply to a privacy request"""

    name = Column(String, unique=True, nullable=False)
    key = Column(String, index=True, unique=True, nullable=False)
    drp_action = Column(EnumColumn(DrpAction), index=True, unique=True, nullable=True)
    execution_timeframe = Column(Integer, nullable=True)
    client_id = Column(
        String,
        ForeignKey(ClientDetail.id_field_path),
        nullable=True,
    )
    client = relationship(
        ClientDetail,
        backref="policies",
    )  # Which client created the Policy

    @classmethod
    def create_or_update(cls, db: Session, *, data: Dict[str, Any]) -> FidesBase:  # type: ignore[override]
        """Overrides base create or update to add custom error for drp action already exists"""
        db_obj = cls.get_by_key_or_id(db=db, data=data)
        if hasattr(cls, "drp_action"):
            data["drp_action"] = data.get("drp_action", None)
            _validate_drp_action(data["drp_action"])
        if db_obj:
            db_obj.update(db=db, data=data)
        else:
            db_obj = cls.create(db=db, data=data)
        return db_obj

    def delete(self, db: Session) -> Optional[FidesBase]:
        """Cascade delete all rules on deletion of a Policy."""
        _ = [rule.delete(db=db) for rule in self.rules]  # type: ignore[attr-defined]
        return super().delete(db=db)

    def get_access_target_categories(self) -> List[str]:
        """Returns all data categories that are the target of access rules."""
        access_categories = []
        for rule in self.rules:  # type: ignore[attr-defined]
            if rule.action_type == ActionType.access:
                access_categories.extend(rule.get_target_data_categories())

        return access_categories

    def get_erasure_target_categories(self) -> List[str]:
        """Returns all data categories that are the target of erasure rules."""
        erasure_categories = []
        for rule in self.rules:  # type: ignore[attr-defined]
            if rule.action_type == ActionType.erasure:
                erasure_categories.extend(rule.get_target_data_categories())

        return erasure_categories

    def get_rules_for_action(self, action_type: ActionType) -> List["Rule"]:
        """Returns all Rules related to this Policy filtered by `action_type`."""
        return [rule for rule in self.rules if rule.action_type == action_type]  # type: ignore[attr-defined]

    def get_consent_rule(self) -> Optional["Rule"]:
        """Returns a Consent Rule if it exists. There should only be one."""
        consent_rules = self.get_rules_for_action(ActionType.consent)
        return consent_rules[0] if consent_rules else None

    def get_action_type(self) -> Optional[ActionType]:
        try:
            return self.rules[0].action_type  # type: ignore[attr-defined]
        except IndexError:
            return None

    def applies_to(self, node: "TraversalNode") -> bool:
        """
        Returns True if any data category in the traversal node starts with any of the policy's target categories.
        """

        target_data_categories: Set[str] = set(
            self.get_access_target_categories()
        ) | set(self.get_erasure_target_categories())

        return any(
            any(category.startswith(target) for target in target_data_categories)
            for category in node.get_data_categories()
        )


def _get_ref_from_taxonomy(
    fides_key: FidesKey,
    all_categories: List[DataCategory] = [],
) -> FideslangDataCategory:
    """Returns the DataCategory model from the DEFAULT_TAXONOMY corresponding to fides_key."""
    if not all_categories:
        all_categories = DEFAULT_TAXONOMY.data_category

    for item in all_categories:
        if item.fides_key == fides_key:
            return item

    raise common_exceptions.DataCategoryNotSupported(
        f"The data category {fides_key} is not configured."
    )


def _is_ancestor_of_contained_categories(
    fides_key: FidesKey,
    data_categories: List[str],
    all_categories: List[DataCategory],
) -> Tuple[bool, Optional[str]]:
    """
    Returns True if `fides_key` is an ancestor of any item in `data_categories`.
    Warning that this algorithm is recursive, is susceptible to infinite loops and
    other misconfigurations in the underlying DEFAULT_TAXONOMY imported from fideslang.

    TODO: Should we memoize this function?
    """
    ref = _get_ref_from_taxonomy(
        fides_key=fides_key,
        all_categories=all_categories,
    )
    if ref.parent_key:
        if ref.parent_key in data_categories:
            return True, ref.parent_key

        return _is_ancestor_of_contained_categories(
            fides_key=ref.parent_key,
            data_categories=data_categories,
            all_categories=all_categories,
        )

    return False, None


def _validate_rule_target_collection(
    db: Session,
    target_categories: List[str],
) -> None:
    """
    Validates that no erasure rules within the Policy have conflicting data category targets:
        - We cannot mask the same data categories multiple times
        - We cannot perform separate masking upon a data category's sub-categories
    """
    all_categories = DataCategory.all(db=db)
    for cat in target_categories:
        # Here we check that `cat` is not an ancestor of any other category within `target_categories`
        is_ancestor, ancestor_fides_key = _is_ancestor_of_contained_categories(
            fides_key=cat,  # type: ignore
            data_categories=target_categories,
            all_categories=all_categories,
        )
        if is_ancestor:
            raise common_exceptions.PolicyValidationError(
                f"Policy rules are invalid, action conflict in erasure rules detected for categories {cat} and {ancestor_fides_key}"
            )


class Rule(Base):
    """
    The constraints to apply to data that matches the RuleTargets of Privacy Requests:
        - How to action the privacy request
        - Where to upload any retrieved data
    """

    name = Column(String, unique=True, nullable=False)
    key = Column(String, index=True, unique=True, nullable=False)
    policy_id = Column(
        String,
        ForeignKey(Policy.id_field_path),
        nullable=False,
    )
    policy = relationship(
        Policy,
        backref="rules",
    )
    action_type = Column(
        EnumColumn(ActionType),
        nullable=False,
    )
    masking_strategy = Column(
        MutableDict.as_mutable(
            StringEncryptedType(
                JSONTypeOverride,
                CONFIG.security.app_encryption_key,
                AesGcmEngine,
                "pkcs5",
            )
        ),
        nullable=True,
    )  # Type bytea in the db
    storage_destination_id = Column(
        String,
        ForeignKey(StorageConfig.id_field_path),
        nullable=True,
    )
    storage_destination = relationship(
        StorageConfig,
        backref="rules",
    )
    client_id = Column(
        String,
        ForeignKey(ClientDetail.id_field_path),
        nullable=True,
    )
    client = relationship(
        ClientDetail,
        backref="rules",
    )  # Which client created the Rule

    def save(self, db: Session) -> FidesBase:
        """Validate this object's data before deferring to the superclass on save"""
        _validate_rule(
            action_type=self.action_type,
            storage_destination_id=self.storage_destination_id,
            masking_strategy=self.masking_strategy,
        )
        return super().save(db=db)

    @classmethod
    def create(cls, db: Session, *, data: Dict[str, Any], check_name: bool = True) -> FidesBase:  # type: ignore[override]
        """Validate this object's data before deferring to the superclass on create"""
        policy_id: Optional[str] = data.get("policy_id")

        if not policy_id:
            raise common_exceptions.RuleValidationError(
                "Policy id must be specified on Rule create."
            )

        policy = Policy.get_by(db=db, field="id", value=policy_id)
        if not policy:
            raise common_exceptions.RuleValidationError(
                "Policy id must be specified on Rule create."
            )
        existing_consent_rules = policy.get_rules_for_action(ActionType.consent)

        if (
            existing_consent_rules
            and data.get("action_type") == ActionType.consent.value
        ):
            raise common_exceptions.RuleValidationError(
                f"Policies can only have one consent rule attached.  Existing rule {existing_consent_rules[0].key} found."
            )
        _validate_rule(
            action_type=data.get("action_type"),
            storage_destination_id=data.get("storage_destination_id"),
            masking_strategy=data.get("masking_strategy"),
        )
        return super().create(db=db, data=data, check_name=check_name)

    def delete(self, db: Session) -> Optional[FidesBase]:
        """Cascade delete all targets on deletion of a Rule."""
        _ = [target.delete(db=db) for target in self.targets]  # type: ignore[attr-defined]
        return super().delete(db=db)

    def get_target_data_categories(self) -> List[str]:
        """
        Returns a list of DataCategory enum values representing the targets
        that this Rule is configured to apply to.
        """
        return [target.data_category for target in self.targets]  # type: ignore[attr-defined]

    def get_storage_destination(self, db: Session) -> StorageConfig:
        """
        Utility to return the appropriate proper storage destination for the Rule.
        If the Rule does not have an explicit `storage_destination` set, then the
        application's default storage config will be returned
        """
        if self.storage_destination:
            return self.storage_destination
        storage_destination = get_active_default_storage_config(db)
        if storage_destination is None:
            raise StorageConfigNotFoundException(
                f"The given rule `{self.key}` has no `storage_destination` configured, and there is no active default storage configuration defined"
            )
        return storage_destination

    @classmethod
    def create_or_update(cls, db: Session, *, data: Dict[str, Any]) -> FidesBase:  # type: ignore[override]
        """
        An override of `FidesBase.create_or_update` that handles the specific edge case where
        a `Rule` getting updated may be having its `policy_id` changed, potentially causing
        `Rule`s to unexpectedly bounce between `Policy`ies.
        """
        db_obj = None
        identifier = None

        if data.get("id") is not None:
            # If `id` has been included in `data`, preference that
            db_obj = cls.get(db=db, object_id=data["id"])
            identifier = data.get("id")
        elif data.get("key") is not None:
            # Otherwise, try with `key`
            db_obj = cls.get_by(db=db, field="key", value=data["key"])
            identifier = data.get("key")

        if db_obj:
            if db_obj.policy_id != data["policy_id"]:
                raise common_exceptions.RuleValidationError(
                    f"Rule with identifier {identifier} belongs to another policy."
                )
            db_obj.update(db=db, data=data)
        else:
            db_obj = cls.create(db=db, data=data)  # type: ignore[assignment]

        return db_obj  # type: ignore[return-value]


def _validate_rule_target_name(name: str) -> None:
    """Raises an error if `name` is None"""
    if not name:
        raise common_exceptions.RuleTargetValidationError(
            "A `name` field must be supplied."
        )


class RuleTarget(Base):
    """Which data categories to apply the referenced Rule to"""

    name = Column(String, unique=True, nullable=False)
    key = Column(String, index=True, unique=True, nullable=False)
    data_category = Column(
        String,
        nullable=False,
    )
    rule_id = Column(
        String,
        ForeignKey(Rule.id_field_path),
        nullable=False,
    )
    rule = relationship(
        Rule,
        backref="targets",
    )
    client_id = Column(
        String,
        ForeignKey(ClientDetail.id_field_path),
        nullable=True,
    )
    client = relationship(
        ClientDetail,
        backref="rule_targets",
    )  # Which client created this RuleTarget

    __table_args__ = (
        # NB. __table_args__ requires a Tuple
        UniqueConstraint("rule_id", "data_category", name="_rule_id_data_category_uc"),
    )

    @classmethod
    def get_compound_key(cls, data: Dict[str, Any]) -> str:
        data_category = data.get("data_category")
        if not data_category:
            raise common_exceptions.RuleTargetValidationError(
                "A data_category must be supplied."
            )
        rule_id = data.get("rule_id")
        if not rule_id:
            raise common_exceptions.RuleTargetValidationError(
                "A rule_id must be supplied."
            )

        return f"{rule_id}-{data_category}"

    @classmethod
    def create_or_update(cls, db: Session, *, data: Dict[str, Any]) -> FidesBase:  # type: ignore[override]
        """
        An override of `FidesBase.create_or_update` that handles the specific edge case where
        a `RuleTarget` getting updated may be having its `rule_id` changed, potentially causing
        `RuleTarget`s to unexpectedly bounce between `Rule`s.
        """
        db_obj = None
        identifier = None

        if data.get("id") is not None:
            # If `id` has been included in `data`, preference that
            db_obj = cls.get(db=db, object_id=data["id"])
            identifier = data.get("id")
        elif data.get("key") is not None:
            # Otherwise, try with `key`
            db_obj = cls.get_by(db=db, field="key", value=data["key"])
            identifier = data.get("key")

        if db_obj:
            if db_obj.rule_id != data["rule_id"]:
                raise common_exceptions.RuleTargetValidationError(
                    f"RuleTarget with identifier {identifier} belongs to another rule."
                )
            db_obj.update(db=db, data=data)
        else:
            db_obj = cls.create(db=db, data=data)  # type: ignore[assignment]

        return db_obj  # type: ignore[return-value]

    @classmethod
    def create(cls, db: Session, *, data: Dict[str, Any], check_name: bool = True) -> FidesBase:  # type: ignore[override]
        """Validate data_category on object creation."""
        data_category = data.get("data_category")
        if not data_category:
            raise common_exceptions.RuleTargetValidationError(
                "A data_category must be supplied."
            )
        rule_id = data.get("rule_id")
        if not rule_id:
            raise common_exceptions.RuleTargetValidationError(
                "A rule_id must be supplied."
            )

        default_name = cls.get_compound_key(data=data)
        if data.get("name") is None:
            data["name"] = default_name

        _validate_data_category(
            db=db,
            data_category=data_category,
        )

        # This database query is necessary since we need to access all Rules and their Targets
        # associated with any given Policy, not just those in the local scope of this object.
        rule = Rule.get(db=db, object_id=rule_id)
        if not rule:
            raise common_exceptions.RuleTargetValidationError(
                f"Rule with ID {rule_id} does not exist."
            )

        if rule.action_type.value == ActionType.erasure.value:  # type: ignore[attr-defined]
            # If we're adding a data category to an erasure rule, we need to validate that there
            # are no conflicting actions in the erasure rules.
            erasure_categories = [data_category]
            policy = rule.policy
            if policy:
                erasure_categories.extend(rule.policy.get_erasure_target_categories())

            _validate_rule_target_collection(db, erasure_categories)

        return super().create(db=db, data=data, check_name=check_name)

    def save(self, db: Session) -> FidesBase:
        """Validate data_category on object save."""
        _validate_data_category(
            db=db,
            data_category=self.data_category,
        )
        _validate_rule_target_name(name=self.name)
        return super().save(db=db)

    def update(self, db: Session, *, data: Dict[str, Any]) -> FidesBase:
        """Validate data_category on object update."""
        updated_data_category = data.get("data_category")
        try:
            name = data["name"]
        except KeyError:
            pass
        else:
            if name is None:
                # Don't pass explcit `None` through for `name` because
                # the field is non-nullable
                del data["name"]

        if (
            updated_data_category is not None
            and updated_data_category != self.data_category
        ):
            _validate_data_category(
                db=db,
                data_category=updated_data_category,
            )
        return super().update(db=db, data=data)


class WebhookDirection(EnumType):
    """The webhook direction"""

    one_way = "one_way"  # No response expected
    two_way = "two_way"  # Response expected


class WebhookBase:
    """Mixin class to contain common fields between PolicyPreWebhooks and PolicyPostWebhooks"""

    @declared_attr
    def __tablename__(cls) -> str:
        return cls.__name__.lower()  # type: ignore

    name = Column(String, unique=True, nullable=False)
    key = Column(String, index=True, unique=True, nullable=False)

    @declared_attr
    def policy_id(cls: "WebhookBase") -> Column:
        """Policy id defined as declared_attr because this is needed for FK's on mixins"""
        return Column(
            String,
            ForeignKey(Policy.id_field_path),
            nullable=False,
        )

    @declared_attr
    def connection_config_id(cls: "WebhookBase") -> Column:
        """Connection config id defined as declared_attr because this is needed for FK's on mixins"""
        return Column(
            String, ForeignKey(ConnectionConfig.id_field_path), nullable=False
        )

    direction = Column(
        EnumColumn(WebhookDirection),
        nullable=False,
    )
    order = Column(Integer, nullable=False)

    def reorder_related_webhooks(self, db: Session, new_index: int) -> None:
        """Updates the order of the current webhook, and order of related webhooks where applicable.

        For example, if you had five Pre-Execution webhooks on a Policy and you updated the order of the
        fifth webhook to be the second webhook, the second, third, fourth, and fifth Pre-Execution
        Webhooks on that Policy would likewise have their order updated.
        """

        cls = self.__class__
        webhooks = getattr(self.policy, f"{cls.prefix}_execution_webhooks").order_by(  # type: ignore
            cls.order
        )  # pylint: disable=W0143

        if new_index > webhooks.count() - 1 or new_index < 0:
            raise WebhookOrderException(
                f"Cannot set order to {new_index}: there are only {webhooks.count()} {cls.__name__}(s) defined on this Policy."
            )
        webhook_order = [webhook.key for webhook in webhooks]
        webhook_order.insert(new_index, webhook_order.pop(self.order))

        for webhook in webhooks:
            webhook.update(db=db, data={"order": webhook_order.index(webhook.key)})
        db.commit()


class PolicyPreWebhook(WebhookBase, Base):
    """
    The configuration to describe webhooks that run before
    Privacy Requests are executed for a given Policy.
    """

    prefix = "pre"  # For logging purposes

    connection_config = relationship(
        ConnectionConfig,
        backref="policy_pre_execution_webhooks",
    )

    policy = relationship(
        "Policy", backref=backref("pre_execution_webhooks", lazy="dynamic")
    )

    @classmethod
    def persist_obj(
        cls, db: Session, resource: "PolicyPreWebhook"
    ) -> "PolicyPreWebhook":
        """Override to have PolicyPreWebhooks not be committed to the database automatically."""
        db.add(resource)
        return resource


class PolicyPostWebhook(WebhookBase, Base):
    """
    The configuration to describe webhooks that run after
    Privacy Requests are executed for a given Policy.
    """

    prefix = "post"  # For logging purposes

    connection_config = relationship(
        ConnectionConfig,
        backref="policy_post_execution_webhooks",
    )

    policy = relationship(
        "Policy", backref=backref("post_execution_webhooks", lazy="dynamic")
    )

    @classmethod
    def persist_obj(
        cls, db: Session, resource: "PolicyPostWebhook"
    ) -> "PolicyPostWebhook":
        """Override to have PolicyPostWebhooks not be committed to the database automatically."""
        db.add(resource)
        return resource


WebhookTypes = Union[PolicyPreWebhook, PolicyPostWebhook]



"""automigrate viewer role

Revision ID: 50180bbbb959
Revises: 68a518a3c050
Create Date: 2023-03-14 14:26:32.910570

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
from sqlalchemy import text

revision = "50180bbbb959"
down_revision = "68a518a3c050"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_constraint(
        "messagingconfig_service_type_key", "messagingconfig", type_="unique"
    )
    op.create_index(
        op.f("ix_messagingconfig_service_type"),
        "messagingconfig",
        ["service_type"],
        unique=True,
    )

    """Automatic data migration: Set user scopes to empty, and set all user roles to viewer"""
    bind = op.get_bind()
    bind.execute(
        text(
            """
            UPDATE fidesuserpermissions 
            SET scopes = '{}', 
                roles = '{viewer}';
        """
        )
    )

    """Automatic data migration: Similarly Update all client scopes attached to users"""
    bind.execute(
        text(
            """
                UPDATE client 
                SET scopes = '{}', 
                    roles = '{viewer}'
                FROM fidesuserpermissions
                WHERE fidesuserpermissions.user_id = client.user_id;
            """
        )
    )


def downgrade():
    op.drop_index(op.f("ix_messagingconfig_service_type"), table_name="messagingconfig")
    op.create_unique_constraint(
        "messagingconfig_service_type_key", "messagingconfig", ["service_type"]
    )



"""Contains the db group of the commands for fides."""

import rich_click as click

from fides.cli.options import yes_flag
from fides.cli.utils import with_analytics, with_server_health_check
from fides.common.utils import echo_red, handle_cli_response
from fides.core import api as _api


@click.group(name="db")
@click.pass_context
def database(ctx: click.Context) -> None:
    """
    Run actions against the application database.
    """


@database.command(name="init", deprecated=True)
@click.pass_context
@with_analytics
@with_server_health_check
def db_init(ctx: click.Context) -> None:
    """
    Runs all upgrade migrations for the Fides database.

    Will also automatically initialize a fresh database.

    **WARNING**: Deprecated, use `upgrade` instead.
    """
    config = ctx.obj["CONFIG"]
    handle_cli_response(
        _api.db_action(
            server_url=config.cli.server_url,
            headers=config.user.auth_header,
            action="upgrade",
        )
    )


@database.command(name="upgrade")
@click.pass_context
@with_analytics
@with_server_health_check
def db_upgrade(ctx: click.Context) -> None:
    """
    Runs all upgrade migrations for the Fides database.

    Will also automatically initialize a fresh database.
    """
    config = ctx.obj["CONFIG"]
    handle_cli_response(
        _api.db_action(
            server_url=config.cli.server_url,
            headers=config.user.auth_header,
            action="upgrade",
        )
    )


@database.command(name="reset")
@click.pass_context
@yes_flag
@with_analytics
@with_server_health_check
def db_reset(ctx: click.Context, yes: bool) -> None:
    """
    Reset the database back to its initial state.
    """
    config = ctx.obj["CONFIG"]
    if yes:
        are_you_sure = "y"
    else:
        echo_red(
            "This will drop all data from the Fides database and reload the default taxonomy!"
        )
        are_you_sure = input("Are you sure [y/n]? ")

    if are_you_sure.lower() == "y":
        handle_cli_response(
            _api.db_action(
                server_url=config.cli.server_url,
                headers=config.user.auth_header,
                action="reset",
            )
        )
    else:
        print("Aborting!")



from datetime import datetime, timezone
from typing import List, Tuple

from fideslang.models import Dataset as FideslangDataset
from loguru import logger
from sqlalchemy import select
from sqlalchemy.orm import Session

from fides.api.schemas.dataset import ValidateDatasetResponse
from fides.service.dataset.dataset_validator import DatasetValidator

from fides.api.models.sql_models import (  # type: ignore[attr-defined] # isort: skip
    Dataset as CtlDataset,
)


class DatasetError(Exception):
    """Base class for dataset-related errors"""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)


class DatasetNotFoundException(DatasetError):
    """Raised when a dataset is not found"""


class DatasetService:
    def __init__(self, db: Session):
        self.db = db

    def validate_dataset(
        self,
        dataset: FideslangDataset,
    ) -> ValidateDatasetResponse:
        """
        Validates a standalone dataset for create/update operations, performing all necessary validations.
        """

        return DatasetValidator(self.db, dataset).validate()

    def create_dataset(self, dataset: FideslangDataset) -> CtlDataset:
        """Create a new dataset with validation"""
        self.validate_dataset(dataset)
        data_dict = dataset.model_dump(mode="json")
        return CtlDataset.create(self.db, data=data_dict)

    def update_dataset(self, dataset: FideslangDataset) -> CtlDataset:
        """Update an existing dataset with validation"""

        self.validate_dataset(dataset)
        existing = _get_ctl_dataset(self.db, dataset.fides_key)
        if not existing:
            raise DatasetNotFoundException(f"Dataset {dataset.fides_key} not found")

        # Update the dataset
        data_dict = dataset.model_dump(mode="json")
        return existing.update(self.db, data=data_dict)

    def get_dataset(self, fides_key: str) -> CtlDataset:
        """Get a single dataset by fides key"""
        dataset = _get_ctl_dataset(self.db, fides_key)
        if not dataset:
            raise DatasetNotFoundException(f"Dataset {fides_key} not found")
        return dataset

    def delete_dataset(self, fides_key: str) -> CtlDataset:
        """Delete a dataset by fides key"""
        dataset = self.get_dataset(fides_key)
        dataset.delete(self.db)
        return dataset

    def upsert_datasets(self, datasets: List[FideslangDataset]) -> Tuple[int, int]:
        """
        For any dataset in `datasets` that already exists in the database,
        update the dataset by its `fides_key`. Otherwise, create a new dataset.

        Returns a tuple of (inserted_count, updated_count).
        """
        inserted = 0
        updated = 0

        for dataset in datasets:
            try:
                existing = (
                    self.db.query(CtlDataset)
                    .filter(CtlDataset.fides_key == dataset.fides_key)
                    .first()
                )

                if existing:
                    self.validate_dataset(dataset)
                    data_dict = dataset.model_dump(mode="json")
                    existing.update(self.db, data=data_dict)
                    updated += 1
                else:
                    self.create_dataset(dataset)
                    inserted += 1
            except Exception as e:
                logger.error(f"Error upserting dataset {dataset.fides_key}: {str(e)}")
                raise

        return inserted, updated

    def clean_datasets(self) -> Tuple[List[str], List[str]]:
        datasets = self.db.execute(select([CtlDataset])).scalars().all()
        return _run_clean_datasets(self.db, datasets)


def _get_ctl_dataset(db: Session, fides_key: str) -> CtlDataset:
    """Helper to get CTL dataset by fides_key"""
    ctl_dataset = db.query(CtlDataset).filter(CtlDataset.fides_key == fides_key).first()
    if not ctl_dataset:
        raise DatasetNotFoundException(
            f"No CTL dataset found with fides_key '{fides_key}'"
        )
    return ctl_dataset


def _run_clean_datasets(
    db: Session, datasets: List[FideslangDataset]
) -> tuple[List[str], List[str]]:
    """
    Clean the dataset name and structure to remove any malformed data possibly present from nested field regressions.
    Changes dot separated positional names to source names (ie. `user.address.street` -> `street`).
    """

    for dataset in datasets:
        logger.info(f"Cleaning field names for dataset: {dataset.fides_key}")
        for collection in dataset.collections:
            collection["fields"] = _recursive_clean_fields(collection["fields"])  # type: ignore # pylint: disable=unsupported-assignment-operation

        # manually upsert the dataset

        logger.info(f"Upserting dataset: {dataset.fides_key}")
        failed = []
        try:
            dataset_ctl_obj = (
                db.query(CtlDataset)
                .filter(CtlDataset.fides_key == dataset.fides_key)
                .first()
            )
            if dataset_ctl_obj:
                db.query(CtlDataset).filter(
                    CtlDataset.fides_key == dataset.fides_key
                ).update(
                    {
                        "collections": dataset.collections,
                        "updated_at": datetime.now(timezone.utc),
                    },
                    synchronize_session=False,
                )
                db.commit()
            else:
                logger.error(f"Dataset with fides_key {dataset.fides_key} not found.")
        except Exception as e:
            logger.error(f"Error upserting dataset: {dataset.fides_key} {e}")
            db.rollback()
            failed.append(dataset.fides_key)

    succeeded = [dataset.fides_key for dataset in datasets]
    return succeeded, failed


def _recursive_clean_fields(fields: List[dict]) -> List[dict]:
    """
    Recursively clean the fields of a dataset.
    """
    cleaned_fields = []
    for field in fields:
        field["name"] = field["name"].split(".")[-1]
        if field["fields"]:
            field["fields"] = _recursive_clean_fields(field["fields"])
        cleaned_fields.append(field)
    return cleaned_fields



"""Contains the scan group of the commands for fides."""

import rich_click as click

from fides.cli.options import (
    aws_access_key_id_option,
    aws_region_option,
    aws_secret_access_key_option,
    aws_session_token_option,
    connection_string_option,
    coverage_threshold_option,
    credentials_id_option,
    manifests_dir_argument,
    okta_org_url_option,
    okta_token_option,
    organization_fides_key_option,
)
from fides.cli.utils import (
    handle_aws_credentials_options,
    handle_database_credentials_options,
    handle_okta_credentials_options,
    with_analytics,
)
from fides.core import dataset as _dataset
from fides.core import system as _system


@click.group(name="scan")
@click.pass_context
def scan(ctx: click.Context) -> None:
    """
    Scan and report on discrepancies between Fides resource files and real infrastructure.
    """


@scan.group(name="dataset")
@click.pass_context
def scan_dataset(ctx: click.Context) -> None:
    """
    Scan and report on Fides Dataset resources.
    """


@scan_dataset.command(name="db")
@click.pass_context
@manifests_dir_argument
@credentials_id_option
@connection_string_option
@coverage_threshold_option
@with_analytics
def scan_dataset_db(
    ctx: click.Context,
    manifests_dir: str,
    connection_string: str,
    credentials_id: str,
    coverage_threshold: int,
) -> None:
    """
    Scan a database directly using a SQLAlchemy-style connection string.

    _If there are fields within the database that aren't listed and categorized
    within one of the datasets, this counts as lacking coverage._
    """
    config = ctx.obj["CONFIG"]
    actual_connection_string = handle_database_credentials_options(
        fides_config=config,
        connection_string=connection_string,
        credentials_id=credentials_id,
    )

    _dataset.scan_dataset_db(
        connection_string=actual_connection_string,
        manifest_dir=manifests_dir,
        coverage_threshold=coverage_threshold,
        url=config.cli.server_url,
        headers=config.user.auth_header,
        local=config.cli.local_mode,
    )


@scan.group(name="system")
@click.pass_context
def scan_system(ctx: click.Context) -> None:
    """
    Scan and report on Fides System resources.
    """


@scan_system.command(name="okta")
@click.pass_context
@manifests_dir_argument
@credentials_id_option
@okta_org_url_option
@okta_token_option
@organization_fides_key_option
@coverage_threshold_option
@with_analytics
def scan_system_okta(
    ctx: click.Context,
    manifests_dir: str,
    credentials_id: str,
    org_url: str,
    token: str,
    org_key: str,
    coverage_threshold: int,
) -> None:
    """
    Scan an Okta account and compare applications with annotated Fides Systems.
    """

    config = ctx.obj["CONFIG"]
    okta_config = handle_okta_credentials_options(
        fides_config=config, token=token, org_url=org_url, credentials_id=credentials_id
    )

    _system.scan_system_okta(
        okta_config=okta_config,
        coverage_threshold=coverage_threshold,
        organization_key=org_key,
        manifest_dir=manifests_dir,
        url=config.cli.server_url,
        headers=config.user.auth_header,
    )


@scan_system.command(name="aws")
@click.pass_context
@manifests_dir_argument
@credentials_id_option
@aws_access_key_id_option
@aws_secret_access_key_option
@aws_session_token_option
@aws_region_option
@organization_fides_key_option
@coverage_threshold_option
@with_analytics
def scan_system_aws(
    ctx: click.Context,
    manifests_dir: str,
    credentials_id: str,
    access_key_id: str,
    secret_access_key: str,
    session_token: str,
    region: str,
    org_key: str,
    coverage_threshold: int,
) -> None:
    """
    Scan an AWS account and compare objects with annotated Fides Systems.

    _Scannable resources: [Redshift, RDS, DynamoDb, S3]_
    """
    config = ctx.obj["CONFIG"]
    aws_config = handle_aws_credentials_options(
        fides_config=config,
        access_key_id=access_key_id,
        secret_access_key=secret_access_key,
        session_token=session_token,
        region=region,
        credentials_id=credentials_id,
    )

    _system.scan_system_aws(
        manifest_dir=manifests_dir,
        organization_key=org_key,
        aws_config=aws_config,
        coverage_threshold=coverage_threshold,
        url=config.cli.server_url,
        headers=config.user.auth_header,
    )



from typing import List

from sqlalchemy.engine import URL, LegacyCursorResult  # type: ignore

from fides.api.graph.execution import ExecutionNode
from fides.api.schemas.connection_configuration import MicrosoftSQLServerSchema
from fides.api.service.connectors.query_configs.microsoft_sql_server_query_config import (
    MicrosoftSQLServerQueryConfig,
)
from fides.api.service.connectors.query_configs.query_config import SQLQueryConfig
from fides.api.service.connectors.sql_connector import SQLConnector
from fides.api.util.collection_util import Row
from fides.config import get_config

CONFIG = get_config()


class MicrosoftSQLServerConnector(SQLConnector):
    """
    Connector specific to Microsoft SQL Server
    """

    secrets_schema = MicrosoftSQLServerSchema

    def build_uri(self) -> URL:
        """
        Build URI of format
        mssql+pymssql://[username]:[password]@[host]:[port]/[dbname]
        Returns URL obj, since SQLAlchemy's create_engine method accepts either a URL obj or a string
        """

        config = self.secrets_schema(**self.configuration.secrets or {})

        url = URL.create(
            "mssql+pymssql",
            username=config.username,
            password=config.password,
            host=config.host,
            port=config.port,
            database=config.dbname,
        )

        return url

    def query_config(self, node: ExecutionNode) -> SQLQueryConfig:
        """Query wrapper corresponding to the input execution_node."""
        return MicrosoftSQLServerQueryConfig(node)

    @staticmethod
    def cursor_result_to_rows(results: LegacyCursorResult) -> List[Row]:
        """
        Convert SQLAlchemy results to a list of dictionaries
        """
        return SQLConnector.default_cursor_result_to_rows(results)



"""Contains the view group of the commands for fides."""

import rich_click as click
from toml import dumps

from fides.cli.utils import with_analytics
from fides.common.utils import echo_red, print_divider
from fides.core.utils import get_credentials_path, read_credentials_file


@click.group(name="view")
@click.pass_context
def view(ctx: click.Context) -> None:
    """
    View various resources types.
    """


@view.command(name="config")  # type: ignore
@click.pass_context
@click.argument("section", default="", type=str)
@click.option(
    "--exclude-unset",
    is_flag=True,
    help="Only print configuration values explicitly set by the user.",
)
@with_analytics
def view_config(
    ctx: click.Context, section: str = "", exclude_unset: bool = False
) -> None:
    """
    Prints the configuration values being used for this command-line instance.

    _Note: To see the configuration values being used by the webserver, `GET` the `/api/v1/config` endpoint._
    """
    config = ctx.obj["CONFIG"]
    config_dict = config.model_dump(exclude_unset=exclude_unset)
    if section:
        config_dict = config_dict[section]

    print_divider()
    print(dumps(config_dict))


@view.command(name="credentials")
@click.pass_context
@with_analytics
def view_credentials(ctx: click.Context) -> None:
    """
    Prints the credentials file.
    """
    credentials_path = get_credentials_path()
    try:
        credentials = read_credentials_file(credentials_path)
    except FileNotFoundError:
        echo_red(f"No credentials file found at path: {credentials_path}")
        raise SystemExit(1)

    print_divider()
    print(dumps(credentials.model_dump(mode="json")))



from typing import Optional

from pydantic import ConfigDict, EmailStr

from fides.api.schemas.base_class import FidesSchema


class GetRegistrationStatusResponse(FidesSchema):
    """
    Reflects the registration status of a Fides deployment.
    """

    opt_in: bool = False
    model_config = ConfigDict(extra="forbid")


class Registration(GetRegistrationStatusResponse):
    """
    Describes a Fides registration.
    """

    analytics_id: str
    opt_in: bool
    user_email: Optional[EmailStr]
    user_organization: Optional[str]



# pylint: disable=too-many-instance-attributes
from __future__ import annotations

from copy import deepcopy
from datetime import datetime
from itertools import product
from typing import Any, Dict, List, Literal, Optional, Tuple, TypeVar
from uuid import uuid4

import pydash
from fideslang.models import FidesDatasetReference
from loguru import logger
from sqlalchemy.orm import Session

from fides.api.common_exceptions import FidesopsException
from fides.api.graph.execution import ExecutionNode
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import (
    PrivacyRequest,
    RequestTask,
    generate_request_task_callback_jwe,
)
from fides.api.schemas.saas.saas_config import (
    Endpoint,
    ReadSaaSRequest,
    SaaSConfig,
    SaaSRequest,
)
from fides.api.schemas.saas.shared_schemas import SaaSRequestParams
from fides.api.service.connectors.query_configs.query_config import QueryConfig
from fides.api.task.refine_target_path import (
    build_refined_target_paths,
    join_detailed_path,
)
from fides.api.util import saas_util
from fides.api.util.collection_util import (
    Row,
    flatten_dict,
    merge_dicts,
    unflatten_dict,
)
from fides.api.util.saas_util import (
    ALL_OBJECT_FIELDS,
    CUSTOM_PRIVACY_REQUEST_FIELDS,
    FIDESOPS_GROUPED_INPUTS,
    ISO_8601_DATETIME,
    MASKED_OBJECT_FIELDS,
    PRIVACY_REQUEST_ID,
    REPLY_TO,
    REPLY_TO_TOKEN,
    UUID,
    get_identities,
)
from fides.common.api.v1.urn_registry import REQUEST_TASK_CALLBACK, V1_URL_PREFIX
from fides.config.config_proxy import ConfigProxy

T = TypeVar("T")


class SaaSQueryConfig(QueryConfig[SaaSRequestParams]):
    """Query config that generates populated SaaS requests for a given collection"""

    def __init__(
        self,
        node: ExecutionNode,
        endpoints: Dict[str, Endpoint],
        secrets: Dict[str, Any],
        data_protection_request: Optional[SaaSRequest] = None,
        privacy_request: Optional[PrivacyRequest] = None,
        request_task: Optional[RequestTask] = None,
    ):
        super().__init__(node)
        self.collection_name = node.address.collection
        self.endpoints = endpoints
        self.secrets = secrets
        self.data_protection_request = data_protection_request
        self.privacy_request = privacy_request
        self.action: Optional[str] = None
        self.current_request: Optional[SaaSRequest] = None
        self.request_task = request_task

    def get_read_requests_by_identity(self) -> List[ReadSaaSRequest]:
        """
        Returns the appropriate request configs based on the current collection and identity
        """
        collection_name = self.node.address.collection

        try:
            requests = self.endpoints[collection_name].requests
        except KeyError:
            logger.error("The '{}' endpoint is not defined", collection_name)
            return []

        if not requests.read:
            return []

        read_requests = (
            requests.read if isinstance(requests.read, list) else [requests.read]
        )
        filtered_requests = self._requests_using_identity(read_requests)
        # return all the requests if none contained an identity reference
        return read_requests if not filtered_requests else filtered_requests

    def _requests_using_identity(
        self, requests: List[ReadSaaSRequest]
    ) -> List[ReadSaaSRequest]:
        """Filters for the requests using the provided identity"""

        return [
            request
            for request in requests
            if any(
                param_value
                for param_value in request.param_values or []
                if param_value.identity in get_identities(self.privacy_request)
            )
        ]

    def get_erasure_request_by_action(
        self, action: Literal["update", "delete"]
    ) -> Optional[SaaSRequest]:
        """
        Returns the appropriate request config based on the
        current collection and preferred erasure action (update or delete)
        """

        collection_name = self.node.address.collection
        request: Optional[SaaSRequest] = getattr(
            self.endpoints[collection_name].requests, action
        )
        if request:
            logger.debug(
                "Found matching endpoint to {} '{}' collection", action, collection_name
            )
        else:
            logger.debug(
                "Unable to find matching endpoint to {} '{}' collection",
                action,
                collection_name,
            )
        return request

    def get_masking_request(self, db: Session) -> Optional[SaaSRequest]:
        """
        Returns a tuple of the preferred action and SaaSRequest to use for masking.
        An update request is preferred, but we can use a gdpr delete endpoint or
        delete endpoint if not MASKING_STRICT.
        """

        update: Optional[SaaSRequest] = self.get_erasure_request_by_action("update")
        gdpr_delete: Optional[SaaSRequest] = None
        delete: Optional[SaaSRequest] = None

        if not ConfigProxy(db).execution.masking_strict:
            gdpr_delete = self.data_protection_request
            delete = self.get_erasure_request_by_action("delete")

        try:
            # Return first viable option
            action_type: str = next(
                action
                for action in [
                    "update" if update else None,
                    "data_protection_request" if gdpr_delete else None,
                    "delete" if delete else None,
                ]
                if action
            )

            # store action name for logging purposes
            self.action = action_type

            logger.debug(
                "Selecting '{}' action to perform masking request for '{}' collection.",
                action_type,
                self.collection_name,
            )
            return next(request for request in [update, gdpr_delete, delete] if request)
        except StopIteration:
            return None

    def generate_param_value_maps(
        self, input_data: Dict[str, List[Any]], read_request: SaaSRequest
    ) -> List[Dict[str, Any]]:
        """
        Prepares and combines input data to generate all possible parameter-value combinations.

        This function processes secrets, grouped inputs, and other input data to create
        a comprehensive list of parameter-value mappings. It generates the Cartesian product
        of all input values, ensuring all possible combinations are considered.
        """
        filtered_secrets = self._filtered_secrets(read_request)
        grouped_inputs_list = input_data.pop(FIDESOPS_GROUPED_INPUTS, None)

        # unpack the inputs
        # list_ids: [[1,2,3]] -> list_ids: [1,2,3]
        for param_value in read_request.param_values or []:
            if param_value.unpack:
                value = param_value.name
                input_data[value] = pydash.flatten(input_data.get(value))

        param_value_maps = []
        # we want to preserve the grouped_input relationships so we take each
        # individual group and generate the product with the ungrouped inputs
        for grouped_inputs in grouped_inputs_list or [{}]:
            param_value_maps.extend(
                self._generate_product_list(
                    input_data, filtered_secrets, grouped_inputs
                )
            )

        return param_value_maps

    def generate_requests(
        self,
        input_data: Dict[str, List[Any]],
        policy: Optional[Policy],
        read_request: SaaSRequest,
    ) -> List[Tuple[SaaSRequestParams, Dict[str, Any]]]:
        """
        Takes the identity and reference values from input_data and combines them
        with the connector_param values in use by the read request to generate
        a list of request params.
        """

        # set the read_request as the current request so it is available in
        # generate_query (conform to the interface for QueryConfig)
        self.current_request = read_request

        request_params = []
        param_value_maps = self.generate_param_value_maps(input_data, read_request)

        for param_value_map in param_value_maps:
            try:
                request_params.append(
                    (
                        self.generate_query(
                            {name: [value] for name, value in param_value_map.items()},
                            policy,
                        ),
                        param_value_map,
                    )
                )
            except ValueError as exc:
                if read_request.skip_missing_param_values:
                    logger.info(
                        "Skipping optional read request on node {}: {}",
                        self.node.address.value,
                        exc,
                    )
                    continue
                raise exc

        return request_params

    def _filtered_secrets(self, current_request: SaaSRequest) -> Dict[str, Any]:
        """Return a filtered map of secrets used by the request"""
        param_names = [
            param_value.connector_param
            for param_value in current_request.param_values or []
        ]

        # add external references used in the request, as their values
        # are stored in secrets
        if current_request.param_values:
            param_names.extend(
                [
                    external_reference
                    for reference_list in [
                        param_value.references
                        for param_value in current_request.param_values
                        if param_value.references
                    ]
                    for external_reference in reference_list
                    if isinstance(
                        external_reference, str
                    )  # str references are external references
                ]
            )
        return {
            name: value for name, value in self.secrets.items() if name in param_names
        }

    @staticmethod
    def _generate_product_list(*args: Dict[str, Any]) -> List[Dict[str, Any]]:
        """
        Accepts a variable number of dicts and produces the product of the values from all the dicts.

        Example:

            `_generate_product_list({ "first": ["a", "b"] }, { "second": ["1", "2", "3"] })`

        Returns:
        ```
            [
                { "first": "a", "second": "1" }
                { "first": "a", "second": "2" }
                { "first": "a", "second": "3" }
                { "first": "b", "second": "1" }
                { "first": "b", "second": "2" }
                { "first": "b", "second": "3" }
            ]
        ```
        """

        merged_dicts = merge_dicts(*args)
        return [
            dict(zip(merged_dicts.keys(), values))
            for values in product(
                *(
                    value if isinstance(value, list) else [value]
                    for value in merged_dicts.values()
                )
            )
        ]

    def generate_query(
        self,
        input_data: Dict[str, List[Any]],
        policy: Optional[Policy],
    ) -> SaaSRequestParams:
        """
        This returns the method, path, header, query, and body params needed to make an API call.
        This is the API equivalent of building the components of a database
        query statement (select statement, where clause, limit, offset, etc.)
        """

        if not self.current_request:
            raise FidesopsException(
                f"The 'read' action is not defined for the '{self.collection_name}' "
                f"endpoint in {self.node.connection_key}"
            )

        # create the source of param values to populate the various placeholders
        # in the path, headers, query_params, and body
        param_values: Dict[str, Any] = {}
        for param_value in self.current_request.param_values or []:
            if param_value.references or param_value.identity:
                input_list = input_data.get(param_value.name)
                if input_list:
                    param_values[param_value.name] = input_list[0]
            elif param_value.connector_param:
                param_values[param_value.name] = pydash.get(
                    input_data, param_value.connector_param
                )[0]

        if self.privacy_request:
            param_values[PRIVACY_REQUEST_ID] = self.privacy_request.id

        custom_privacy_request_fields = input_data.get(CUSTOM_PRIVACY_REQUEST_FIELDS)
        if (
            isinstance(custom_privacy_request_fields, list)
            and len(custom_privacy_request_fields) > 0
        ):
            param_values[CUSTOM_PRIVACY_REQUEST_FIELDS] = custom_privacy_request_fields[
                0
            ]

        param_values[UUID] = str(uuid4())
        param_values[ISO_8601_DATETIME] = datetime.now().date().isoformat()
        if self.request_task and self.request_task.id:
            param_values[REPLY_TO_TOKEN] = generate_request_task_callback_jwe(
                self.request_task
            )
            param_values[REPLY_TO] = V1_URL_PREFIX + REQUEST_TASK_CALLBACK

        # map param values to placeholders in path, headers, and query params
        saas_request_params: SaaSRequestParams = saas_util.map_param_values(
            self.action, self.collection_name, self.current_request, param_values  # type: ignore
        )

        logger.debug("Populated request params for {}", self.current_request.path)

        return saas_request_params

    def generate_update_stmt(
        self, row: Row, policy: Policy, request: PrivacyRequest
    ) -> SaaSRequestParams:
        """
        This returns the method, path, header, query, and body params needed to make an API call.
        The fields in the row are masked according to the policy and added to the request body
        if specified by the body field of the masking request.
        """
        session = Session.object_session(request)
        current_request: SaaSRequest = self.get_masking_request(session)  # type: ignore
        param_values: Dict[str, Any] = self.generate_update_param_values(
            row, policy, request, current_request
        )

        return self.generate_update_request_params(param_values, current_request)

    def generate_consent_stmt(
        self,
        policy: Policy,
        privacy_request: PrivacyRequest,
        consent_request: SaaSRequest,
    ) -> SaaSRequestParams:
        """
        Prepares SaaSRequestParams with the info needed to make an opt-out or opt-in http request.
        Shares a lot of code with generate_update_stmt, except there is no row data being operated on,
        so our row is an empty dict.
        """

        param_values: Dict[str, Any] = self.generate_update_param_values(
            {}, policy, privacy_request, consent_request
        )

        return self.generate_update_request_params(param_values, consent_request)

    def generate_update_param_values(  # pylint: disable=R0914
        self,
        row: Row,
        policy: Policy,
        privacy_request: PrivacyRequest,
        saas_request: SaaSRequest,
    ) -> Dict[str, Any]:
        """
        A utility that generates the update request param values
        based on the provided inputs for the given SaaSRequest.

        The update param values are returned as a `dict`. The
        `masked_object_fields` key maps to a JSON structure that holds
        the fields in the provided row that have been masked according
        to provided policy. The `all_object_fields` key maps to a JSON structure
        that holds all values, including those that have not been masked.
        """

        collection_name: str = self.node.address.collection
        collection_values: Dict[str, Row] = {collection_name: row}
        identity_data: Dict[str, Any] = privacy_request.get_cached_identity_data()
        custom_privacy_request_fields: Dict[str, Any] = (
            privacy_request.get_cached_custom_privacy_request_fields()
        )

        # create the source of param values to populate the various placeholders
        # in the path, headers, query_params, and body
        param_values: Dict[str, Any] = {}
        for param_value in saas_request.param_values or []:
            if param_value.references:
                # we resolve the param reference here for consistency,
                # i.e. as if it may be a pointer to an `external_reference`.
                # however, `references` in update requests can, currently, only reference
                # the same collection the same collection, and so it is highly unlikely
                # that this would be an external reference at this point.
                reference: FidesDatasetReference = SaaSConfig.resolve_param_reference(
                    param_value.references[0], self.secrets
                )
                param_values[param_value.name] = pydash.get(
                    collection_values, reference.field
                )
            elif param_value.identity:
                param_values[param_value.name] = pydash.get(
                    identity_data, param_value.identity
                )
            elif param_value.connector_param:
                param_values[param_value.name] = pydash.get(
                    self.secrets, param_value.connector_param
                )

        if self.privacy_request:
            param_values[PRIVACY_REQUEST_ID] = self.privacy_request.id

        param_values[CUSTOM_PRIVACY_REQUEST_FIELDS] = custom_privacy_request_fields
        param_values[UUID] = str(uuid4())
        param_values[ISO_8601_DATETIME] = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
        if self.request_task and self.request_task.id:
            param_values[REPLY_TO_TOKEN] = generate_request_task_callback_jwe(
                self.request_task
            )
            param_values[REPLY_TO] = V1_URL_PREFIX + REQUEST_TASK_CALLBACK
        # remove any row values for fields marked as read-only, these will be omitted from all update maps
        for field_path, field in self.field_map().items():
            if field.read_only:
                for path in build_refined_target_paths(
                    row, query_paths={field_path: None}
                ):
                    detailed_path = join_detailed_path(path)
                    pydash.unset(row, detailed_path)

        # mask row values
        update_value_map: Dict[str, Any] = self.update_value_map(
            row, policy, privacy_request
        )
        masked_object: Dict[str, Any] = unflatten_dict(update_value_map)

        # map of all values including those not being masked/updated
        all_value_map: Dict[str, Any] = self.all_value_map(row)
        # both maps use field paths for the keys so we can merge them before unflattening
        # values in update_value_map will override values in all_value_map
        complete_object: Dict[str, Any] = unflatten_dict(
            merge_dicts(all_value_map, update_value_map)
        )

        param_values[MASKED_OBJECT_FIELDS] = masked_object
        param_values[ALL_OBJECT_FIELDS] = complete_object

        return param_values

    def generate_update_request_params(
        self, param_values: dict[str, Any], update_request: SaaSRequest
    ) -> SaaSRequestParams:
        """
        A utility that, based on the provided param values and update request,
        generates the `SaaSRequestParams` that are to be used in request execution
        """

        # map param values to placeholders in path, headers, and query params
        saas_request_params: SaaSRequestParams = saas_util.map_param_values(
            self.action, self.collection_name, update_request, param_values  # type: ignore
        )

        logger.debug("Populated request params for {}", update_request.path)
        return saas_request_params

    def all_value_map(self, row: Row) -> Dict[str, Any]:
        """
        Takes a row and preserves only the fields that are defined in the collection.
        Used for scenarios when an update endpoint has required fields other than
        just the fields being updated.
        """
        flattened_row = flatten_dict(deepcopy(row))

        # Get root field names defined in the collection
        collection_fields = {
            field_path.string_path.split(".")[0]
            for field_path, _ in self.field_map().items()
        }

        # Only keep the field values defined in the collection
        return {
            path: value
            for path, value in flattened_row.items()
            if path.split(".")[0] in collection_fields
        }

    def query_to_str(self, t: T, input_data: Dict[str, List[Any]]) -> str:
        """Convert query to string"""
        return "Not yet supported for SaaSQueryConfig"

    def dry_run_query(self) -> Optional[str]:
        """dry run query for display"""
        return None



from typing import Optional

from sqlalchemy import Column, ForeignKey, Index, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Session

from fides.api.db.base_class import Base
from fides.api.models.fides_user import FidesUser
from fides.config import get_config

CONFIG = get_config()


class SystemHistory(Base):
    """History of changes to a system"""

    @declared_attr
    def __tablename__(self) -> str:
        return "plus_system_history"

    user_id = Column(String, nullable=True)
    system_id = Column(
        String, ForeignKey("ctl_systems.id", ondelete="cascade"), nullable=False
    )
    before = Column(MutableDict.as_mutable(JSONB), nullable=False)
    after = Column(MutableDict.as_mutable(JSONB), nullable=False)

    __table_args__ = (
        Index(
            "idx_plus_system_history_created_at_system_id", "created_at", "system_id"
        ),
    )

    @property
    def edited_by(self) -> Optional[str]:
        """Derive the username from the user_id"""
        if not self.user_id:
            return None

        if self.user_id == CONFIG.security.oauth_root_client_id:
            return CONFIG.security.root_username

        db = Session.object_session(self)
        user: Optional[FidesUser] = FidesUser.get_by(db, field="id", value=self.user_id)

        return user.username if user else self.user_id



# pylint: disable=too-many-lines
from typing import Any, Dict, List, Optional, Tuple

from loguru import logger

from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.service.connectors.query_configs.query_config import QueryConfig
from fides.api.util.collection_util import Row, filter_nonempty_values

MongoStatement = Tuple[Dict[str, Any], Dict[str, Any]]
"""A mongo query is expressed in the form of 2 dicts, the first of which represents
  the query object(s) and the second of which represents fields to return.
  e.g. 'collection.find({k1:v1, k2:v2},{f1:1, f2:1 ... })'. This is returned as
  a tuple ({k1:v1, k2:v2},{f1:1, f2:1 ... }).

  An update statement takes the form
  collection.update_one({k1:v1},{k2:v2}...}, {$set: {f1:fv1, f2:fv2 ... }}, upsert=False).
  This is returned as a tuple
  ({k1:v1},{k2:v2}...},  {f1:fv1, f2: fv2 ... }
  """


class MongoQueryConfig(QueryConfig[MongoStatement]):
    """Query config that translates parameters into mongo statements"""

    def generate_query(
        self, input_data: Dict[str, List[Any]], policy: Optional[Policy] = None
    ) -> Optional[MongoStatement]:
        def transform_query_pairs(pairs: Dict[str, Any]) -> Dict[str, Any]:
            """Since we want to do an 'OR' match in mongo, transform queries of the form
            {A:1, B:2} => "{$or:[{A:1},{B:2}]}".
            Don't bother to do this if the pairs size is 1
            """
            if len(pairs) < 2:
                return pairs
            return {"$or": [dict([(k, v)]) for k, v in pairs.items()]}

        if input_data:
            filtered_data: Dict[str, Any] = self.node.typed_filtered_values(input_data)
            if filtered_data:
                query_pairs = {}
                for string_field_path, data in filtered_data.items():
                    if len(data) == 1:
                        query_pairs[string_field_path] = data[0]

                    elif len(data) > 1:
                        query_pairs[string_field_path] = {"$in": data}

                field_list = {  # Get top-level fields to avoid path collisions
                    field_path.string_path: 1
                    for field_path, field in self.top_level_field_map().items()
                }
                query_fields, return_fields = (
                    transform_query_pairs(query_pairs),
                    field_list,
                )
                return query_fields, return_fields

        logger.warning(
            "There is not enough data to generate a valid query for {}",
            self.node.address,
        )
        return None

    def generate_update_stmt(
        self, row: Row, policy: Policy, request: PrivacyRequest
    ) -> Optional[MongoStatement]:
        """Generate a SQL update statement in the form of Mongo update statement components"""
        update_clauses = self.update_value_map(row, policy, request)

        pk_clauses: Dict[str, Any] = filter_nonempty_values(
            {
                field_path.string_path: field.cast(row[field_path.string_path])
                for field_path, field in self.primary_key_field_paths.items()
            }
        )

        valid = len(pk_clauses) > 0 and len(update_clauses) > 0
        if not valid:
            logger.warning(
                "There is not enough data to generate a valid update for {}",
                self.node.address,
            )
            return None
        return pk_clauses, {"$set": update_clauses}

    def query_to_str(self, t: MongoStatement, input_data: Dict[str, List[Any]]) -> str:
        """string representation of a query for logging/dry-run"""
        query_data, field_list = t
        db_name = self.node.address.dataset
        collection_name = self.node.address.collection
        return f"db.{db_name}.{collection_name}.find({query_data}, {field_list})"

    def dry_run_query(self) -> Optional[str]:
        data = self.display_query_data()
        mongo_query = self.generate_query(self.display_query_data(), None)
        if mongo_query is not None:
            return self.query_to_str(mongo_query, data)
        return None



from __future__ import annotations

import abc
from typing import Any, ClassVar, Dict, List, Optional

from pydantic import BaseModel, ConfigDict, model_validator

from fides.api.models.connectionconfig import ConnectionTestStatus
from fides.api.schemas import Msg


class ConnectionConfigSecretsSchema(BaseModel, abc.ABC):
    """Abstract Base Schema for updating Connection Configuration Secrets"""

    # NOTE: any fields not listed in `_required_components` must also have an
    # annotated Optional type, in order to be treated effectively as optional fields
    # Further, annotating this as as a ClassVar prevents this from being converted to a ModelPrivateAttrs
    _required_components: ClassVar[List[str]]

    def __init_subclass__(cls: BaseModel, **kwargs: Any):  # type: ignore
        super().__init_subclass__(**kwargs)  # type: ignore
        if not getattr(cls, "_required_components"):
            raise TypeError(f"Class {cls.__name__} must define '_required_components.'")  # type: ignore

    @model_validator(mode="before")
    @classmethod
    def required_components_supplied(cls, values) -> Dict[str, Any]:  # type: ignore
        """Validate that the minimum required components have been supplied."""

        min_fields_present = all(
            values.get(component) for component in cls._required_components
        )
        if not min_fields_present:
            raise ValueError(
                f"{cls.__name__} must be supplied all of: {cls._required_components}."  # type: ignore
            )

        return values

    model_config = ConfigDict(extra="ignore", from_attributes=True)


class TestStatusMessage(Msg):
    """A schema for checking status."""

    test_status: Optional[ConnectionTestStatus] = None
    failure_reason: Optional[str] = None



"""tcf_purpose_overrides

Revision ID: 848a8f4125cf
Revises: 7f7c2b098f5d
Create Date: 2023-12-02 13:18:55.768697

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "848a8f4125cf"
down_revision = "7f7c2b098f5d"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "tcf_purpose_overrides",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("purpose", sa.Integer(), nullable=False),
        sa.Column("is_included", sa.Boolean(), server_default="t", nullable=True),
        sa.Column("required_legal_basis", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_tcf_purpose_overrides_id"),
        "tcf_purpose_overrides",
        ["id"],
        unique=False,
    )
    op.create_unique_constraint(
        "purpose_constraint", "tcf_purpose_overrides", ["purpose"]
    )

    op.create_index(
        op.f("ix_tcf_purpose_overrides_purpose"),
        "tcf_purpose_overrides",
        ["purpose"],
        unique=False,
    )


def downgrade():
    op.drop_index(
        op.f("ix_tcf_purpose_overrides_purpose"), table_name="tcf_purpose_overrides"
    )
    op.drop_index(
        op.f("ix_tcf_purpose_overrides_id"), table_name="tcf_purpose_overrides"
    )
    op.drop_table("tcf_purpose_overrides")



"""case_insensitive_usernames_delete_duplicates

Revision ID: 643249f65453
Revises: 5d62bab40b71
Create Date: 2023-02-01 22:22:38.055862

"""

from alembic import op
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "643249f65453"
down_revision = "5d62bab40b71"
branch_labels = None
depends_on = None


def upgrade():
    session = Session(bind=op.get_bind())
    conn = session.connection()
    statement = text(
        """SELECT DISTINCT ON (LOWER(username)) username, created_at, id
FROM fidesuser
ORDER BY LOWER(username), created_at ASC"""
    )

    result = conn.execute(statement)
    original_user_ids = [row[2] for row in result]
    op.execute(
        f"UPDATE client SET user_id = NULL WHERE NOT(user_id=ANY(ARRAY[{original_user_ids}]::varchar[]))"
    )
    op.execute(
        f"UPDATE auditlog SET user_id = 'user_deleted' WHERE NOT(user_id=ANY(ARRAY[{original_user_ids}]::varchar[]))"
    )
    op.execute(
        f"""DELETE FROM fidesuserpermissions
    WHERE NOT(user_id=ANY(ARRAY[{original_user_ids}]::varchar[]))"""
    )

    op.execute(
        f"""DELETE FROM fidesuser 
    WHERE NOT(id=ANY(ARRAY[{original_user_ids}]::varchar[]))"""
    )

    op.execute("CREATE EXTENSION IF NOT EXISTS citext;")
    op.execute("ALTER TABLE fidesuser ALTER COLUMN username TYPE citext;")


def downgrade():
    op.execute("ALTER TABLE fidesuser ALTER COLUMN username TYPE varchar;")
    op.execute("DROP EXTENSION IF EXISTS citext;")



from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field, ValidationError, model_validator
from sqlalchemy import Column
from sqlalchemy import Enum as EnumColumn
from sqlalchemy import ForeignKey, Index, Integer, String, insert, select, update
from sqlalchemy.dialects.postgresql import ARRAY, JSONB
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session

from fides.api.db.base_class import Base


class TCFRestrictionType(str, Enum):
    """Enum for TCF restriction types"""

    purpose_restriction = "purpose_restriction"
    require_consent = "require_consent"
    require_legitimate_interest = "require_legitimate_interest"


class TCFVendorRestriction(str, Enum):
    """Enum for TCF vendor restriction types"""

    restrict_all_vendors = "restrict_all_vendors"
    allow_specific_vendors = "allow_specific_vendors"
    restrict_specific_vendors = "restrict_specific_vendors"


# This is Pydantic model used for validation, not a database model!
class RangeEntry(BaseModel):
    """
    Pydantic model that represents a vendor range entry as per the TCF spec,
    used for Publisher Restrictions.
    A range entry must have a start_vendor_id and optionally an end_vendor_id.
    If end_vendor_id is present, it must be greater than start_vendor_id.
    """

    start_vendor_id: int = Field(description="The starting vendor ID in the range")
    end_vendor_id: Optional[int] = Field(
        default=None, description="The ending vendor ID in the range (inclusive)"
    )

    @model_validator(mode="after")
    def validate_vendor_range(self) -> "RangeEntry":
        """Validates that end_vendor_id is greater than start_vendor_id if present."""
        if (
            self.end_vendor_id is not None
            and self.end_vendor_id <= self.start_vendor_id
        ):
            raise ValueError("end_vendor_id must be greater than start_vendor_id")
        return self

    def get_end(self) -> int:
        """Get the effective end of the range."""
        return self.end_vendor_id or self.start_vendor_id

    def overlaps_with(self, other: "RangeEntry") -> bool:
        """
        Check if this range overlaps with another range.
        Two ranges overlap if the end of the range that starts first is greater than
        or equal to the start of the range that starts second.
        """
        # Sort ranges by start_vendor_id
        first = self if self.start_vendor_id <= other.start_vendor_id else other
        second = other if self.start_vendor_id <= other.start_vendor_id else self

        # If first range's end is >= second range's start, they overlap
        return first.get_end() >= second.start_vendor_id


class TCFConfiguration(Base):
    """
    Stores TCF Configuration settings.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "tcf_configuration"

    name = Column(String, nullable=False, index=True, unique=True)


class TCFPublisherRestriction(Base):
    """
    Stores TCF Publisher Restrictions. TCF Publisher Restrictions belong to a TCF Configuration,
    and specify the restriction type and vendor restriction for a given purpose.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "tcf_publisher_restriction"

    tcf_configuration_id = Column(
        String(255),
        ForeignKey("tcf_configuration.id", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    purpose_id = Column(Integer, nullable=False)
    restriction_type = Column(EnumColumn(TCFRestrictionType), nullable=False)
    vendor_restriction = Column(EnumColumn(TCFVendorRestriction), nullable=False)

    # range_entries represents a list of RangeEntry objects as per the TCF spec.
    # A RangeEntry object will have a start_vendor_id and an end_vendor_id (the
    # end vendor id is included in the range).
    # If the range represents a single vendor, then the end_vendor_id can be omitted.
    # TCF spec also includes an IsARange boolean, which we are omitting because it
    # can be inferred from the presence of the end_vendor_id.
    range_entries = Column(
        ARRAY(JSONB),
        nullable=True,
        server_default="{}",
        default=list,
    )

    __table_args__ = (
        Index(
            "ix_tcf_publisher_restriction_config_purpose",
            "tcf_configuration_id",
            "purpose_id",
        ),
    )

    def __init__(self, **kwargs: Any) -> None:
        validated_kwargs = self.validate_publisher_restriction_data(kwargs)
        super().__init__(**validated_kwargs)

    @staticmethod
    def validate_entires_for_vendor_restriction(
        entries: List[RangeEntry], vendor_restriction: TCFVendorRestriction
    ) -> None:
        """
        Validates that if vendor_restriction is restrict_all_vendors, then the entries list is empty.
        """
        if vendor_restriction == TCFVendorRestriction.restrict_all_vendors and entries:
            raise ValueError("restrict_all_vendors cannot have any range entries")

    @staticmethod
    def check_for_overlaps(entries: List[RangeEntry]) -> None:
        """
        Check for overlapping ranges in a list of RangeEntry objects,
        raises a ValueError if any pair of RangeEntry overlaps with each other.
        Sorts the ranges by start_vendor_id and compares adjacent ranges.
        """
        # Sort ranges by start_vendor_id
        sorted_entries = sorted(entries, key=lambda x: x.start_vendor_id)

        # Compare each range with the next one
        for i in range(len(sorted_entries) - 1):
            current = sorted_entries[i]
            next_entry = sorted_entries[i + 1]
            if current.overlaps_with(next_entry):
                raise ValueError(
                    f"Overlapping ranges found: {current.model_dump()} overlaps with {next_entry.model_dump()}"
                )

    @classmethod
    def validate_publisher_restriction_data(
        cls, data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Validate the restriction data.
        """
        raw_range_entries = data.get("range_entries", [])

        try:
            # Validate each range entry using the Pydantic model
            validated_entries = [
                RangeEntry.model_validate(entry) for entry in raw_range_entries
            ]
        except ValidationError as e:
            raise ValueError(f"Invalid range entry: {str(e)}")

        # Validate the entries for the vendor restriction
        cls.validate_entires_for_vendor_restriction(
            validated_entries, data["vendor_restriction"]
        )
        # Check for overlapping ranges
        cls.check_for_overlaps(validated_entries)

        data["range_entries"] = [entry.model_dump() for entry in validated_entries]

        return data

    @classmethod
    async def validate_vendor_overlaps_for_purpose(
        cls,
        async_db: AsyncSession,
        configuration_id: str,
        purpose_id: int,
        new_data: dict,
    ) -> None:
        """
        Validates that the new vendor ranges do not overlap with any existing vendor ranges for the purpose
        in the given configuration.
        Raises a ValueError if any vendor ranges overlap.
        """
        # First, get all the restrictions for the purpose in the given configuration
        query = (
            select(cls)  # type: ignore[arg-type]
            .where(cls.tcf_configuration_id == configuration_id)
            .where(cls.purpose_id == purpose_id)
        )
        restrictions = await async_db.execute(query)
        restrictions = restrictions.scalars().all()

        # If we have an existing id, we need to exclude the current restriction from the list of existing restrictions
        # so that the new_data restrictions don't overlap with themselves
        if "id" in new_data:
            existing_entries = [
                entry
                for r in restrictions
                for entry in r.range_entries
                if r.id != new_data["id"]
            ]
        else:
            existing_entries = [
                entry for r in restrictions for entry in r.range_entries
            ]

        all_entries = [*existing_entries, *new_data.get("range_entries", [])]
        cls.check_for_overlaps(
            [RangeEntry.model_validate(entry) for entry in all_entries]
        )

    @classmethod
    def create(
        cls,
        db: Session,
        *,
        data: Dict[str, Any],
        check_name: bool = True,
    ) -> "TCFPublisherRestriction":
        raise NotImplementedError("Use create_async instead")

    @classmethod
    async def create_async(
        cls,
        async_db: AsyncSession,
        *,
        data: Dict[str, Any],
    ) -> "TCFPublisherRestriction":
        """
        Create a new TCFPublisherRestriction with validated range_entries.
        """

        data = cls.validate_publisher_restriction_data(data)

        values = {
            "tcf_configuration_id": data["tcf_configuration_id"],
            "purpose_id": data["purpose_id"],
            "restriction_type": data["restriction_type"],
            "vendor_restriction": data["vendor_restriction"],
            "range_entries": data["range_entries"],
        }

        # Validate that the new vendor ranges do not overlap with any existing vendor ranges for the purpose
        await cls.validate_vendor_overlaps_for_purpose(
            async_db=async_db,
            configuration_id=data["tcf_configuration_id"],
            purpose_id=data["purpose_id"],
            new_data=values,
        )

        # Insert the new restriction
        insert_stmt = insert(cls).values(values)  # type: ignore[arg-type]
        result = await async_db.execute(insert_stmt)
        record_id = result.inserted_primary_key.id

        created_record = await async_db.execute(select(cls).where(cls.id == record_id))  # type: ignore[arg-type]
        return created_record.scalars().first()

    async def update_async(
        self, async_db: AsyncSession, data: Dict[str, Any]
    ) -> "TCFPublisherRestriction":
        """
        Update a TCFPublisherRestriction with the data.
        Validates the data and checks for vendor overlaps.
        """
        # Validate the data on its own
        data = self.validate_publisher_restriction_data(data)

        # Validate that the new vendor ranges do not overlap with any existing vendor ranges for the purpose
        await self.validate_vendor_overlaps_for_purpose(
            async_db=async_db,
            configuration_id=self.tcf_configuration_id,
            purpose_id=self.purpose_id,
            new_data={**data, "id": self.id},  # Pass in id explicitly
        )

        # Finally, make the update
        update_query = (
            update(TCFPublisherRestriction)  # type: ignore[arg-type]
            .where(TCFPublisherRestriction.id == self.id)
            .values(**data)
        )
        await async_db.execute(update_query)
        await async_db.commit()
        await async_db.refresh(self)

        return self



from typing import Dict, List, Literal, Optional

from pydantic import BaseModel, Field

from fides.api.models.consent_automation import ConsentableItem as ConsentableItemModel
from fides.api.models.privacy_notice import UserConsentPreference
from fides.api.schemas.base_class import FidesSchema


class ConsentableItem(FidesSchema):
    """
    Schema to represent 3rd-party consentable items and privacy notice relationships.
    """

    external_id: str
    type: str
    name: str
    notice_id: Optional[str] = None
    children: List["ConsentableItem"] = Field(default_factory=list)
    unmapped: Optional[bool] = False


def merge_consentable_items(
    api_items: List[ConsentableItem], db_items: Optional[List[ConsentableItem]] = None
) -> List[ConsentableItem]:
    """
    Recursively merges the lists of consentable items, setting the unmapped flag if the item is not in the database.

    WARNING: This is a destructive operation for the api_items parameter.
    """

    if db_items is None:
        db_items = []

    def merge_consentable_items_recursive(
        source: ConsentableItem, target: Optional[ConsentableItem]
    ) -> None:
        if target is None:
            source.unmapped = True
            for child in source.children:
                child.unmapped = True
            return

        if source.external_id == target.external_id:
            source.unmapped = False
            source.notice_id = target.notice_id
            target_children_map = {
                child.external_id: child for child in target.children
            }

            for child in source.children:
                target_child = target_children_map.get(child.external_id)
                if target_child is not None:
                    merge_consentable_items_recursive(child, target_child)

    # create a map of target items for efficient lookup
    db_item_map = {item.external_id: item for item in db_items}

    # iterate through API items and merge
    for api_item in api_items:
        target_item = db_item_map.get(api_item.external_id)
        merge_consentable_items_recursive(api_item, target_item)

    return api_items


def build_consent_item_hierarchy(
    consentable_items: List[ConsentableItemModel],
) -> List[ConsentableItem]:
    """
    Builds a hierarchy of ConsentableItem Pydantic models from the flat list of database models.
    """

    return [
        ConsentableItem.model_validate(item)
        for item in consentable_items
        if item.parent_id is None
    ]


class ConsentWebhookResult(BaseModel):
    """
    A wrapper class for the identity map and notice map values returned from a `PROCESS_CONSENT_WEBHOOK` function.
    """

    identity_map: Dict[
        Literal["email", "phone_number", "fides_user_device", "external_id"], str
    ] = {}
    notice_map: Dict[str, UserConsentPreference] = {}

    @property
    def success(self) -> bool:
        """Returns true if both the identity map and notice map are not empty."""
        return bool(self.identity_map) and bool(self.notice_map)



"""Add data migration to reactivate taxonomy nodes

Revision ID: ae65da77c468
Revises: 10c6b7709be3
Create Date: 2024-12-18 12:27:47.239489

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy import text
from sqlalchemy.engine import Connection
from sqlalchemy.sql.elements import TextClause

# revision identifiers, used by Alembic.
revision = "ae65da77c468"
down_revision = "10c6b7709be3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """
    This is a data migration to activate taxonomy nodes if it is de-active and has any children that are active

    E.g.
    Taxonomy Tree: A----B----C
                         \
                          ----D
    Current Active Fields: A (false), B (false), C (true), D (false)
    Upgrade Active Fields: A (true), B (true), C (true), D (false)
    """
    bind: Connection = op.get_bind()

    reactivate_data_categories_query: TextClause = text(
        """
        WITH RECURSIVE leaf_nodes AS (
            SELECT DISTINCT dc.fides_key
            FROM ctl_data_categories dc
            WHERE dc.fides_key NOT IN (
                SELECT DISTINCT parent_key 
                FROM ctl_data_categories 
                WHERE parent_key IS NOT NULL
            )
            AND dc.active = true
        ),
        parent_hierarchy AS (
            SELECT dc.fides_key, dc.parent_key
            FROM ctl_data_categories dc
            INNER JOIN leaf_nodes ln ON dc.fides_key = ln.fides_key
            
            UNION
            
            SELECT dc.fides_key, dc.parent_key
            FROM ctl_data_categories dc
            INNER JOIN parent_hierarchy ph ON dc.fides_key = ph.parent_key
        )
        UPDATE ctl_data_categories
        SET active = true
        WHERE fides_key IN (
            SELECT fides_key FROM parent_hierarchy
        )
        AND active = false;
        """
    )

    reactivate_data_uses_query: TextClause = text(
        """
        WITH RECURSIVE leaf_nodes AS (
            SELECT DISTINCT dc.fides_key
            FROM ctl_data_uses dc
            WHERE dc.fides_key NOT IN (
                SELECT DISTINCT parent_key 
                FROM ctl_data_uses 
                WHERE parent_key IS NOT NULL
            )
            AND dc.active = true
        ),
        parent_hierarchy AS (
            SELECT dc.fides_key, dc.parent_key
            FROM ctl_data_uses dc
            INNER JOIN leaf_nodes ln ON dc.fides_key = ln.fides_key
            
            UNION
            
            SELECT dc.fides_key, dc.parent_key
            FROM ctl_data_uses dc
            INNER JOIN parent_hierarchy ph ON dc.fides_key = ph.parent_key
        )
        UPDATE ctl_data_uses
        SET active = true
        WHERE fides_key IN (
            SELECT fides_key FROM parent_hierarchy
        )
        AND active = false;
        """
    )

    # Update ctl_data_categories
    bind.execute(reactivate_data_categories_query)

    # Update ctl_data_uses
    bind.execute(reactivate_data_uses_query)


def downgrade() -> None:
    """
    This migration does not support downgrades.
    """
    pass



from loguru import logger
from sqlalchemy import text
from sqlalchemy.orm import Session

from fides.api.api.deps import get_autoclose_db_session
from fides.api.db.base_class import FidesBase
from fides.api.migrations.hash_migration_mixin import HashMigrationMixin
from fides.api.migrations.hash_migration_tracker import HashMigrationTracker
from fides.api.models.privacy_preference import (
    CurrentPrivacyPreference,
    PrivacyPreferenceHistory,
    ServedNoticeHistory,
)
from fides.api.models.privacy_request import CustomPrivacyRequestField, ProvidedIdentity
from fides.api.tasks.scheduled.scheduler import scheduler
from fides.config import CONFIG

HASH_MIGRATION = "hash_migration"
MODELS_TO_MIGRATE = [
    CurrentPrivacyPreference,
    ProvidedIdentity,
    CustomPrivacyRequestField,
    PrivacyPreferenceHistory,
    ServedNoticeHistory,
]


def initiate_bcrypt_migration_task() -> None:
    """Initiates scheduler to migrate all non-credential tables using a bcrypt hash to use a SHA-256 hash"""

    if CONFIG.test_mode:
        return

    assert (
        scheduler.running
    ), "Scheduler is not running! Cannot migrate tables with bcrypt hashes."

    logger.info("Initiating scheduler for hash migration")
    scheduler.add_job(
        func=bcrypt_migration_task,
        id=HASH_MIGRATION,
    )


def bcrypt_migration_task() -> None:
    """
    Job to migrate all the tables using bcrypt hashes for general data (excludes tables with credentials).
    """

    with get_autoclose_db_session() as db:
        # Do a single pass to check if any of the models have already been migrated.
        # This will allow us to optimize searching for these models by not calling
        # the previously used bcrypt hash.
        for model in MODELS_TO_MIGRATE:
            if is_migrated(db, model):
                HashMigrationTracker.set_migrated(model.__name__)
                logger.info(f"{model.__name__} is already fully migrated.")

        # migrate remaining models
        for model in MODELS_TO_MIGRATE:
            if not HashMigrationTracker.is_migrated(model.__name__):
                migrate_table(db, model)  # type: ignore[arg-type]


def is_migrated(db: Session, model: type[FidesBase]) -> bool:
    """
    Checks the database to see if there exists at least one unmigrated row.
    """

    query = text(
        f"SELECT EXISTS (SELECT 1 FROM {model.__tablename__} WHERE is_hash_migrated IS FALSE)"
    )
    result = db.execute(query).scalar()
    return not result


def migrate_table(
    db: Session, model: type[HashMigrationMixin], batch_size: int = 1000
) -> None:
    """
    Migrate all unmigrated rows in the table corresponding to the passed in model.
    """

    while True:
        unmigrated_batch = (
            db.query(model)
            .filter(model.is_hash_migrated.is_(False))
            .limit(batch_size)
            .all()
        )
        if not unmigrated_batch:
            break

        for item in unmigrated_batch:
            item.migrate_hashed_fields()

        # commit after each batch is complete
        db.commit()

    HashMigrationTracker.set_migrated(model.__name__)
    logger.info(f"Completed hash migration for {model.__name__}.")



from fides.api.service.masking.strategy import (
    masking_strategy_aes_encrypt,
    masking_strategy_hash,
    masking_strategy_hmac,
    masking_strategy_nullify,
    masking_strategy_random_string_rewrite,
    masking_strategy_string_rewrite,
)



from fides.api.schemas.base_class import FidesSchema


class IdentityVerificationConfigResponse(FidesSchema):
    """Response for identity verification config info"""

    identity_verification_required: bool
    disable_consent_identity_verification: bool
    valid_email_config_exists: bool



"""create privacy declarations table

Revision ID: 48d9caacebd4
Revises: 3842d1acac5f
Create Date: 2023-04-20 20:35:05.377471

"""

import json
import uuid
from collections import defaultdict

import sqlalchemy as sa
from alembic import op
from sqlalchemy import text
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "48d9caacebd4"
down_revision = "3842d1acac5f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "privacydeclaration",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column("egress", sa.ARRAY(sa.String()), nullable=True),
        sa.Column("ingress", sa.ARRAY(sa.String()), nullable=True),
        sa.Column("data_use", sa.String(), nullable=False),
        sa.Column("data_categories", sa.ARRAY(sa.String()), nullable=True),
        sa.Column("data_qualifier", sa.String(), nullable=True),
        sa.Column("data_subjects", sa.ARRAY(sa.String()), nullable=True),
        sa.Column("dataset_references", sa.ARRAY(sa.String()), nullable=True),
        sa.Column("system_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["system_id"],
            ["ctl_systems.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )

    op.create_index(
        op.f("ix_privacydeclaration_data_use"),
        "privacydeclaration",
        ["data_use"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacydeclaration_id"), "privacydeclaration", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_privacydeclaration_name"), "privacydeclaration", ["name"], unique=False
    )
    op.create_index(
        op.f("ix_privacydeclaration_system_id"),
        "privacydeclaration",
        ["system_id"],
        unique=False,
    )

    # Data migration

    bind = op.get_bind()
    existing_declarations = bind.execute(
        text("SELECT id, privacy_declarations FROM ctl_systems;")
    )
    for row in existing_declarations:
        system_id = row["id"]
        old_privacy_declarations = row["privacy_declarations"]
        for privacy_declaration in old_privacy_declarations:
            new_privacy_declaration_id: str = "pri_" + str(uuid.uuid4())
            new_data = {
                **privacy_declaration,
                "system_id": system_id,
                "id": new_privacy_declaration_id,
            }

            insert_privacy_declarations_query = text(
                "INSERT INTO privacydeclaration (id, name, data_categories, data_qualifier, data_subjects, dataset_references, egress, ingress, system_id, data_use) "
                "VALUES (:id, :name, :data_categories, :data_qualifier, :data_subjects, :dataset_references, :egress, :ingress, :system_id, :data_use)"
            )

            bind.execute(
                insert_privacy_declarations_query,
                new_data,
            )

    op.drop_column("ctl_systems", "privacy_declarations")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "ctl_systems",
        sa.Column(
            "privacy_declarations",
            postgresql.JSON(astext_type=sa.Text()),
            autoincrement=False,
            nullable=True,
        ),
    )

    # Data migration

    bind = op.get_bind()
    existing_declarations = bind.execute(
        text(
            "SELECT id, name, data_categories, data_qualifier, data_subjects, dataset_references, egress, ingress, system_id, data_use FROM privacydeclaration;"
        )
    )
    pds_by_system = defaultdict(list)
    for row in existing_declarations:
        system_id = row["system_id"]
        privacy_declaration = {
            "name": row["name"],
            "data_categories": row["data_categories"],
            "data_qualifier": row["data_qualifier"],
            "data_subjects": row["data_subjects"],
            "dataset_references": row["dataset_references"],
            "egress": row["egress"],
            "ingress": row["ingress"],
            "data_use": row["data_use"],
        }
        pds_by_system[system_id].append(privacy_declaration)

    for system_id, pds in pds_by_system.items():
        privacy_declarations = json.dumps(pds)
        update_systems_query = text(
            "update ctl_systems set privacy_declarations = :privacy_declarations where id = :system_id;"
        )

        bind.execute(
            update_systems_query,
            {"system_id": system_id, "privacy_declarations": privacy_declarations},
        )

    op.drop_index(
        op.f("ix_privacydeclaration_system_id"), table_name="privacydeclaration"
    )
    op.drop_index(op.f("ix_privacydeclaration_name"), table_name="privacydeclaration")
    op.drop_index(op.f("ix_privacydeclaration_id"), table_name="privacydeclaration")
    op.drop_index(
        op.f("ix_privacydeclaration_data_use"), table_name="privacydeclaration"
    )
    op.drop_table("privacydeclaration")
    # ### end Alembic commands ###



from __future__ import annotations

from sqlalchemy import Column, Enum, String, Text
from sqlalchemy.ext.declarative import declared_attr

from fides.api.db.base_class import Base
from fides.api.schemas.custom_asset import CustomAssetType


class CustomAsset(Base):
    @declared_attr
    def __tablename__(self) -> str:
        return "plus_custom_asset"

    key = Column(
        Enum(CustomAssetType),
        index=True,
        nullable=False,
    )
    filename = Column(String, nullable=False)
    content = Column(Text, nullable=False)



"""update_special_purpose_unique_constraint

Revision ID: 4cb3b5af4160
Revises: 8c54e1e5bdc4
Create Date: 2023-09-27 22:58:43.064996

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "4cb3b5af4160"
down_revision = "8c54e1e5bdc4"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_constraint(
        "identity_special_purpose", "currentprivacypreference", type_="unique"
    )
    op.create_unique_constraint(
        "identity_special_purpose",
        "currentprivacypreference",
        ["provided_identity_id", "special_purpose"],
    )


def downgrade():
    op.drop_constraint(
        "identity_special_purpose", "currentprivacypreference", type_="unique"
    )
    op.create_unique_constraint(
        "identity_special_purpose",
        "currentprivacypreference",
        ["provided_identity_id", "purpose"],
    )



"""add_reject_all_mechanism_to_privacy_experience_config

Revision ID: 67d01c4e124e
Revises: 1d2f04ac19f2
Create Date: 2025-03-27 15:26:24.635947

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "67d01c4e124e"
down_revision = "1d2f04ac19f2"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    # Add the reject all mechanism enum and column
    op.execute(
        "CREATE TYPE rejectallmechanism AS ENUM ('REJECT_ALL', 'REJECT_CONSENT_ONLY')"
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "reject_all_mechanism",
            sa.Enum("REJECT_ALL", "REJECT_CONSENT_ONLY", name="rejectallmechanism"),
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column(
            "reject_all_mechanism",
            sa.Enum("REJECT_ALL", "REJECT_CONSENT_ONLY", name="rejectallmechanism"),
            nullable=True,
        ),
    )

    # Data migration - set the reject_all_mechanism for all TCF experience configs to REJECT_ALL
    op.execute(
        "UPDATE privacyexperienceconfig SET reject_all_mechanism = 'REJECT_ALL' WHERE component = 'tcf_overlay'"
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("privacyexperienceconfig", "reject_all_mechanism")
    op.drop_column("privacyexperienceconfighistory", "reject_all_mechanism")
    op.execute("DROP TYPE IF EXISTS rejectallmechanism")
    # ### end Alembic commands ###



"""Add error message tracking

Revision ID: 8f84fad4e00b
Revises: 58933b5cc6e8
Create Date: 2022-11-15 01:38:28.531640

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8f84fad4e00b"
down_revision = "58933b5cc6e8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "privacyrequesterror",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("message_sent", sa.Boolean(), nullable=False),
        sa.Column("privacy_request_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["privacy_request_id"],
            ["privacyrequest.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_privacyrequesterror_id"), "privacyrequesterror", ["id"], unique=False
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_privacyrequesterror_id"), table_name="privacyrequesterror")
    op.drop_table("privacyrequesterror")
    # ### end Alembic commands ###



"""add consent request relationship

Revision ID: 61a922702f4c
Revises: 5a8cee9c014c
Create Date: 2024-01-04 22:57:24.155581

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "61a922702f4c"
down_revision = "5a8cee9c014c"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "custom_privacy_request_field",
        sa.Column("consent_request_id", sa.String(), nullable=True),
    )
    op.create_foreign_key(
        "consent_request_id_fkey",
        "custom_privacy_request_field",
        "consentrequest",
        ["consent_request_id"],
        ["id"],
    )


def downgrade():
    op.drop_constraint(
        "consent_request_id_fkey", "custom_privacy_request_field", type_="foreignkey"
    )
    op.drop_column("custom_privacy_request_field", "consent_request_id")
